'use strict';

var vendor_js = require('../dist/vendor.js');
var events = require('events');
var path = require('path');
var fs = require('fs');

const log = vendor_js.debug('bundler:file-utils');

const SLASH = '/'.charCodeAt(0);

function isDirectory (p) {
  return (fs.existsSync(p) && fs.lstatSync(p).isDirectory())
}

function isAbsolute (p) {
  // this is broken since node 8
  // return path.resolve(p) === path.normalize(p)
  return path.isAbsolute(p)
}

function copySync (src, dest) {
  const dir = _dir(dest);
  log(`ensureDirSync(${dir})`);
  vendor_js.fse.ensureDirSync(dir);
  vendor_js.fse.copySync(src, dest);
}

function writeSync (dest, buf) {
  let dir = _dir(dest);
  vendor_js.fse.ensureDirSync(dir);
  fs.writeFileSync(dest, buf);
}

function mkDirSync (dir) {
  vendor_js.fse.ensureDirSync(dir);
}

function rmSync (rmPath) {
  if (fs.existsSync(rmPath)) {
    vendor_js.fse.removeSync(rmPath);
  }
}

function _dir (p) {
  if (p.charCodeAt(p.length - 1) === SLASH) return p
  else return path.dirname(p)
}

var isWindows = (process.platform === 'win32');

const log$1 = vendor_js.debug('bundler:file-watcher');

class FileWatcher extends events.EventEmitter {
  constructor (file) {
    super();

    const opts = {
      encoding: 'utf-8',
      persistent: true
    };
    // TODO: a good setting of awaitWriteFinish prevents double updates
    if (isWindows) {
      opts.awaitWriteFinish = true;
    } else {
      opts.awaitWriteFinish = {
        stabilityThreshold: 100,
        pollInterval: 50
      };
    }

    this._file = file;
    const watcher = vendor_js.chokidar.watch(file, opts);
    this._watcher = watcher;
    watcher.setMaxListeners(0);
    watcher.on('error', function (err) {
      if (err.code === 'ENOENT') ; else {
        throw err
      }
    });
    watcher.on('unlink', this.onDelete.bind(this));
    watcher.on('change', this.onChange.bind(this));
  }

  close () {
    this._watcher.close();
  }

  onDelete (file) {
    log$1('File deleted: %s', file);
    this.emit('unlink', this._file);
    // TODO: we need to think about a way to explicitly free watchers
    // ATM, watchers are only freed when stopping the bundler
    // this.close()
  }

  onChange (file) {
    log$1('File changed: %s', file);
    this.emit('change', file);
  }
}

const opts = {
  encoding: 'utf-8',
  persistent: true,
  ignoreInitial: true
};

class GlobWatcher {
  constructor (pattern) {
    const watcher = vendor_js.chokidar.watch(pattern, opts);
    this._watcher = watcher;
    watcher.setMaxListeners(0);
    watcher.on('error', function (err) {
      throw err
    });
  }

  close () {
    this._watcher.close();
  }

  on (evt, handler) {
    this._watcher.on(evt, handler);
  }
}

class Watcher {
  constructor () {
    this.watching = false;
    this.watchEntries = {};
    this.watchers = {};

    // make sure that watchers are disposed on exit
    this.stop = this.stop.bind(this);
    process.on('SIGINT', this.stop);
    process.on('exit', this.stop);
  }

  start () {
    if (this.watching) return
    let watchEntries = this.watchEntries;
    let watchers = this.watchers;
    Object.keys(watchEntries).forEach(function (id) {
      let watcher = watchers[id];
      if (watcher) watcher.close();
      let watchEntry = watchEntries[id];
      watchers[id] = this._createWatcher(watchEntry);
    }.bind(this));
    this.watching = true;
  }

  stop () {
    let watchEntries = this.watchEntries;
    let watchers = this.watchers;
    Object.keys(watchEntries).forEach(function (id) {
      let watcher = watchers[id];
      if (watcher) {
        watcher.close();
        delete watchers[id];
      }
    });
    this.watching = false;
  }

  watch (patternOrPath, hooks) {
    if (!patternOrPath || !vendor_js.isString(patternOrPath)) {
      throw new Error('Watcher.watch(): invalid path ' + patternOrPath)
    }
    const watchEntries = this.watchEntries;
    const watchers = this.watchers;
    let watchEntry;
    let type = vendor_js.glob.hasMagic(patternOrPath) ? 'glob' : 'file';
    if (type === 'file' && !isAbsolute(patternOrPath)) {
      throw new Error('Only absolute file paths are allowed. Was ' + patternOrPath)
    }
    if (!watchEntries[patternOrPath]) {
      watchEntry = {
        type: type,
        path: patternOrPath,
        handlers: [hooks]
      };
      watchEntries[patternOrPath] = watchEntry;
    } else {
      watchEntries[patternOrPath].handlers.push(hooks);
      return
    }
    if (this.watching) {
      watchers[patternOrPath] = this._createWatcher(watchEntry);
    }
  }

  _createWatcher (watchEntry) {
    const w = watchEntry.type === 'file'
      ? new FileWatcher(watchEntry.path)
      : new GlobWatcher(watchEntry.path);
    // console.log('### watching %s', watchEntry.path)
    watchEntry.handlers.forEach(function (hooks) {
      Object.keys(hooks).forEach(function (evt) {
        w.on(evt, hooks[evt]);
      });
    });
    return w
  }
}

class Task {
  constructor (name, fn, deps) {
    this.name = name;
    this.fn = fn;
    this.deps = deps;
    this.description = '';
  }
  describe (description) {
    this.description = description;
    return this
  }
}

const log$2 = vendor_js.debug('bundler:action');

class Action {
  constructor (inputs, outputs) {
    this.inputs = inputs || [];
    this.outputs = outputs || [];
  }

  invalidate () {
    // nothing by default
    // Note: some actions use Action.removeOutputs()
  }

  get id () {
    return String(this.inputs) + '->' + String(this.outputs)
  }

  get descr () {
    return this.id
  }

  _removeOutputs () {
    Action.removeOutputs(this);
  }

  static removeOutputs (action) {
    action.outputs.forEach(function (file) {
      log$2('Invalidating %s', file);
      vendor_js.fse.removeSync(file);
    });
  }
}

const log$3 = vendor_js.debug('bundler:copy');

/*
  Copy a single file into dist folder:
  ```
    copy('./foo.js', 'dist/')
  ```

  Copy a single file with renaming
  ```
    copy('./foo.js', 'dist/bar.js')
  ```

  Copy a whole directory into dist folder
  ```
    copy('./assets', 'dist/')
  ```

  Copy a whole directory into dist folder with renaming
  ```
    copy('./node_modules/substance/dist', 'dist/substance')
  ```

  Copy with glob pattern
  ```
    copy('./node_modules/substance/packages/** /*.css', 'dist/styles/', { root: './node_modules/substance/'})
  ```
*/

class CopyCommand {
  constructor (src, dest, opts) {
    this.src = src;
    this.dest = dest;
    this.opts = opts || {};
  }

  get id () {
    return ['CopyCommand', this.src, '->', this.dest].join(' ')
  }

  get name () {
    return 'copy'
  }

  execute (bundler) {
    bundler._info('Copy:', this.src, '->', this.dest);
    const rootDir = bundler.rootDir;
    let dest = this.dest;
    if (!isAbsolute(dest)) dest = path.join(rootDir, dest);
    const lastIsSlash = (dest[dest.length - 1] === path.sep);
    if (vendor_js.glob.hasMagic(this.src)) {
      // use the parent of the first * as relative root
      // for paths found via glob
      if (!this.opts.root) {
        let idx = this.src.lastIndexOf('*');
        let root = path.join(rootDir, path.dirname(this.src.slice(0, idx + 1)));
        // console.log('USING ROOT', root)
        this.opts.root = root;
      }
      return this._executeWithGlob(bundler)
    } else if (isDirectory(this.src)) {
      log$3('Copying a directory...');
      if (this.opts.root) ; else if (lastIsSlash) {
        if (isAbsolute(this.src)) {
          this.opts.root = path.dirname(this.src);
        } else {
          this.opts.root = path.join(rootDir, path.dirname(this.src));
        }
        log$3('using implicit root dir: ' + this.opts.root);
      } else {
        this.opts.root = this.src;
        log$3('using implicit root dir: ' + this.opts.root);
      }
      this.src = this.src + '/**/*';
      return this._executeWithGlob(bundler)
    } else {
      return this._executeWithoutGlob(bundler)
    }
  }

  _executeWithoutGlob (bundler) {
    const rootDir = bundler.rootDir;
    let dest = this.dest;
    let src = this.src;
    if (!isAbsolute(src)) src = path.join(rootDir, src);
    if (!isAbsolute(dest)) dest = path.join(rootDir, dest);
    const lastIsSlash = (dest[dest.length - 1] === path.sep);
    if (lastIsSlash || isDirectory(dest)) dest = path.join(dest, path.basename(src));
    // console.log('####', dest, isAbsolute(dest), dest[dest.length-1], path.sep, lastIsSlash, path.basename(src))
    const action = new CopyAction(src, dest);
    bundler._registerAction(action);
    action._execute();
  }

  _executeWithGlob (bundler) {
    const rootDir = bundler.rootDir;
    const watcher = bundler.watcher;
    const pattern = this.src;
    let dest = this.dest;
    if (!isAbsolute(dest)) dest = path.join(rootDir, dest);
    let globRoot = rootDir;
    if (this.opts.root) {
      if (isAbsolute(this.opts.root)) {
        globRoot = this.opts.root;
      } else {
        globRoot = path.join(rootDir, this.opts.root);
      }
    }
    // console.log('GLOBROOT', globRoot)
    const files = vendor_js.glob.sync(pattern);
    if (files) {
      files.forEach(function (relativePath) {
        // console.log('RELATIVE_PATH', relativePath)
        let absPath = path.join(rootDir, relativePath);
        // console.log('ABS_PATH', absPath)
        let srcPath = absPath;
        if (isDirectory(srcPath)) return
        let destPath = path.join(dest, path.relative(globRoot, absPath));
        // console.log('DEST_PATH', destPath)
        const action = new CopyAction(srcPath, destPath);
        bundler._registerAction(action);
        action._execute();
      });
    } else {
      console.error('No files found for pattern %s', pattern);
    }

    // add a watcher that is used to register actions
    // for added files
    watcher.watch(pattern, {
      'add': this._onAdd.bind(this, bundler)
    });
  }

  /*
    When used with a glob pattern this gets called
    when a new file has been added
  */
  _onAdd (bundler, file) {
    if (isDirectory(file)) return
    const rootDir = bundler.rootDir;
    let srcPath = file;
    if (!isAbsolute(srcPath)) srcPath = path.join(rootDir, srcPath);
    if (!fs.existsSync(srcPath)) return
    const globRoot = this.opts.root || rootDir;
    let destPath = path.join(this.dest, path.relative(globRoot, file));
    if (!isAbsolute(destPath)) destPath = path.join(rootDir, destPath);
    const action = new CopyAction(srcPath, destPath);
    bundler._registerAction(action);
    bundler._schedule(action);
  }
}

class CopyAction extends Action {
  constructor (src, dest) {
    super([src], [dest]);

    if (!isAbsolute(src)) {
      throw new Error('Only absolute paths are allowed. Was ' + src)
    }
    if (!isAbsolute(dest)) {
      throw new Error('Only absolute paths are allowed. Was ' + dest)
    }
    // use real paths so that we do not register copy actions
    // for the same file twice (~ in presence of sym links)
    src = fs.realpathSync(src);

    this.src = src;
    this.dest = dest;
  }

  get id () {
    return ['Copy:', this.src, '->', this.dest].join(' ')
  }

  execute (bundler) {
    bundler._info(this.id);
    this._execute();
  }

  invalidate () {
    vendor_js.fse.removeSync(this.dest);
  }

  _execute () {
    const src = this.src;
    const dest = this.dest;
    if (fs.existsSync(src)) {
      log$3(`copying ${src} -> ${dest}`);
      copySync(src, dest);
    }
  }
}

function randomId () {
  const length = 8;
  const ts = Date.now().toString();
  const parts = ts.split('').reverse();
  const id = new Array(length);
  for (let i = length - 1; i >= 0; i--) {
    id[i] = _getRandomInt(0, parts.length - 1);
  }
  return id.join('')
}

function _getRandomInt (min, max) {
  return Math.floor(Math.random() * (max - min + 1)) + min
}

class CustomCommand {
  constructor (description, params) {
    this._id = randomId();
    this.description = description;
    this.src = params.src;
    this.dest = params.dest;
    this._execute = params.execute;
    if (!vendor_js.isFunction(this._execute)) {
      throw new Error("'execute' must be a function")
    }
  }

  get id () {
    return this._id
  }

  get descr () {
    return this.description
  }

  get name () {
    return 'custom'
  }

  execute (bundler) {
    if (this.src && vendor_js.isArray(this.src)) {
      return this._executeWithCollection(bundler)
    } else if (this.src && vendor_js.glob.hasMagic(this.src)) {
      return this._executeWithGlob(bundler)
    } else {
      return this._executeWithoutGlob(bundler)
    }
  }

  _executeWithoutGlob (bundler) {
    let src = this.src;
    let dest = this._normalizedDest(bundler);
    if (src && !isAbsolute(src)) src = path.join(bundler.rootDir, src);
    const action = new CustomAction(this._id, this.description, [src], dest, this._execute, false);
    bundler._registerAction(action);
    return action.execute(bundler)
  }

  _executeWithGlob (bundler) {
    const rootDir = bundler.rootDir;
    const watcher = bundler.watcher;
    const pattern = this.src;
    let dest = this._normalizedDest(bundler);
    let files = vendor_js.glob.sync(pattern, {});
    if (files) {
      files = files.map(function (file) {
        return path.join(rootDir, file)
      });
      const action = new CustomAction(this._id, this.description, files, dest, this._execute, true);
      bundler._registerAction(action);
      let result = action.execute(bundler);
      // TODO: need to rework the whole dynamic registry stuff
      watcher.watch(pattern, {
        add: function (file) {
          action.inputs.push(file);
          action.inputs = vendor_js.uniq(action.inputs);
          const _actionsByInput = bundler._actionsByInput;
          if (!_actionsByInput[file]) _actionsByInput[file] = [];
          _actionsByInput[file] = vendor_js.uniq(_actionsByInput[file].concat(action));
        }
      });
      return result
    } else {
      console.error('No files found for pattern %s', pattern);
    }
  }

  _executeWithCollection (bundler) {
    const rootDir = bundler.rootDir;
    const watcher = bundler.watcher;
    let dest = this._normalizedDest(bundler);
    let files = [];
    let patterns = [];
    this.src.forEach(function (fileOrPattern) {
      if (vendor_js.glob.hasMagic(fileOrPattern)) {
        patterns.push(fileOrPattern);
        files = files.concat(vendor_js.glob.sync(fileOrPattern, {}));
      } else {
        files.push(fileOrPattern);
      }
    });
    if (files.length) {
      files = files.map(function (file) {
        return path.join(rootDir, file)
      });
      const action = new CustomAction(this._id, this.description, files, dest, this._execute, true);
      bundler._registerAction(action);
      let result = action.execute(bundler);
      patterns.forEach(function (pattern) {
        // TODO: need to rework the whole dynamic registry stuff
        watcher.watch(pattern, {
          add: function (file) {
            action.inputs.push(file);
            action.inputs = vendor_js.uniq(action.inputs);
            const _actionsByInput = bundler._actionsByInput;
            if (!_actionsByInput[file]) _actionsByInput[file] = [];
            _actionsByInput[file] = vendor_js.uniq(_actionsByInput[file].concat(action));
          }
        });
      });
      return result
    } else {
      console.error('No files found for pattern %s', patterns);
    }
  }

  _normalizedDest (bundler) {
    const rootDir = bundler.rootDir;
    let dest = this.dest;
    if (dest) {
      if (!vendor_js.isArray(dest)) dest = [dest];
      dest = dest.map(d => {
        if (d && !isAbsolute(d)) d = path.join(rootDir, d);
        return d
      });
    } else {
      dest = [];
    }
    return dest
  }
}

class CustomAction extends Action {
  constructor (id, description, inputs, outputs, _execute, multipleSourceFiles) {
    super(inputs.filter(Boolean), outputs.filter(Boolean));
    this._id = id;
    this._description = description;
    this._execute = _execute;
    this._watched = new Set();
    this._multipleSourceFiles = multipleSourceFiles;
  }

  get id () {
    return this._id
  }

  get description () {
    return this._description
  }

  execute (bundler) {
    bundler._info(this._description);
    let t0 = Date.now();
    this.outputs.forEach(function (f) {
      vendor_js.fse.ensureDirSync(path.dirname(f));
    });
    let inputs = this.inputs.filter(function (file) {
      return fs.existsSync(file)
    });
    let api = {
      watch: (file) => {
        if (!isAbsolute(file)) file = path.join(bundler.rootDir, file);
        this._addWatcher(bundler, file);
      },
      // providing some fs api
      isAbsolute,
      isDirectory,
      copySync,
      mkDirSync,
      rmSync,
      writeFileSync: writeSync
    };
    if (!this._multipleSourceFiles) {
      inputs = inputs[0];
    }
    return Promise.resolve(this._execute(inputs, api))
      .then(function () {
        bundler._info(vendor_js.colors.green('..finished in %s ms.'), Date.now() - t0);
      })
  }

  _addWatcher (bundler, absPath) {
    const watcher = bundler.watcher;
    const _onChange = () => {
      _invalidate();
      bundler._schedule(this);
    };
    const _invalidate = () => {
      this.invalidate();
      for (let output of this.outputs) {
        bundler._invalidate(output);
      }
    };
    if (!this._watched.has(absPath)) {
      watcher.watch(absPath, {
        change: _onChange,
        unlink: _invalidate
      });
      this._watched.add(absPath);
    }
  }
}

class MakeCommand {
  constructor (module, tasks) {
    this.module = module;
    this.tasks = tasks || [];
  }

  get id () {
    return ['Make', this.module].concat(this.tasks).join(' ')
  }

  get name () {
    return 'make'
  }

  execute (bundler) {
    bundler._info(this.id);

    const tasks = this.tasks;
    const module = this.module;
    let makefile;
    if (isAbsolute(module)) {
      makefile = module;
    } else {
      makefile = require.resolve(module + '/make.js');
    }
    if (!fs.existsSync(makefile)) {
      throw new Error('Could not find "make.js" in module "%s"', module)
    }
    makefile = fs.realpathSync(makefile);

    return new Promise(_runMake)

    function _runMake (resolve, reject) {
      // get the real location of the module in case
      // it is npm-linked
      const cp = require('child_process');
      let args = tasks.concat(['--remote']);
      if (bundler.opts.watch) args.push('--watch');
      const child = cp.fork(makefile, args, {
        cwd: path.dirname(makefile)
      });
      child.on('message', function (msg) {
        if (msg === 'done') {
          resolve();
        }
      });
      child.on('error', function (error) {
        reject(new Error(error));
      });
    }
  }
}

class RemoveCommand {
  constructor (rmPath) {
    this.rmPath = rmPath;
  }

  get id () {
    return ['Remove:', this.rmPath].join(' ')
  }

  get name () {
    return 'rm'
  }

  execute (bundler) {
    let rmPath = this.rmPath;
    if (rmPath[0] === '.') {
      rmPath = path.join(bundler.rootDir, rmPath);
    }
    if (fs.existsSync(rmPath)) {
      bundler._info(this.id);
      vendor_js.fse.removeSync(this.rmPath);
    }
  }
}

const log$4 = vendor_js.debug('bundler');

class Bundler extends events.EventEmitter {
  constructor (opts) {
    super();

    this.opts = opts || {};
    this.rootDir = process.cwd();

    this.watcher = new Watcher();
    this.server = new vendor_js.express(); // eslint-disable-line new-cap
    this._serverPort = 4000;

    // an action is registerd via a file id
    // whenever the file changes the action gets scheduled
    this._actions = {};
    this._actionsByInput = {};
    this._generatedFiles = {};

    // a task is just a function executing commands
    // or registering actions
    this._tasks = {};

    // a queue for sequencing actions
    this._scheduledActions = [];
    this._scheduledActionIds = {};

    // flags
    this._started = false;
    this._running = false;
    this._firstRun = true;

    this._step = this._step.bind(this);

    process.on('SIGINT', function () {
      log$4('received SIGINT');
      process.exit(0);
    });
  }

  make (other, ...tasks) {
    return this._scheduleCommand(new MakeCommand(other, tasks))
  }

  copy (src, dest, opts) {
    return this._scheduleCommand(new CopyCommand(src, dest, opts))
  }

  cp (...args) {
    return this.copy(...args)
  }

  custom (description, params) {
    return this._scheduleCommand(new CustomCommand(description, params))
  }

  rm (rmPath) {
    return this._scheduleCommand(new RemoveCommand(rmPath))
  }

  task (name, deps, fn) {
    log$4('Registering task', name);
    if (vendor_js.isFunction(deps)) {
      fn = deps;
      deps = [];
    }
    const task = new Task(name, fn, deps);
    this._tasks[name] = task;
    return task
  }

  setServerPort (port) {
    this._serverPort = port;
  }

  serve (params) {
    const server = this.server;
    const rootDir = this.rootDir;
    if (params.static) {
      const route = params.route;
      let folder = params.folder;
      if (!route || !folder) {
        throw new Error("Parameters 'route' and 'folder' required.")
      }
      folder = path.join(rootDir, folder);
      log$4('Adding static route %s -> %s', route, folder);
      server.use(route, vendor_js.express.static(folder));
    }
  }

  writeFileSync (dest, buf) {
    writeSync(dest, buf);
  }

  writeSync (dest, buf) {
    console.warn('DEPRECATED: use b.writeFileSync()');
    this.writeFileSync(dest, buf);
  }

  isDirectory (p) {
    return isDirectory(p)
  }

  get colors () {
    return vendor_js.colors
  }

  _scheduleCommand (cmd) {
    const entry = this._schedule(cmd);
    return entry.promise
  }

  _info (...args) {
    if (!this.silent) {
      console.info.apply(console, args);
    }
  }

  _hasScheduledActions () {
    return this._scheduledActions.length > 0
  }

  _registerAction (action) {
    const id = action.id;
    // skip if the same action is already registered
    if (this._actions[id]) return
    this._actions[id] = action;
    action.inputs.forEach((input) => {
      this._registerActionInput(action, input);
    });
    action.outputs.forEach((output) => {
      this._registerActionOutput(action, output);
    });
  }

  _registerActionInput (action, input) {
    const watcher = this.watcher;
    // if no other action has been registered as generator
    // then we add a file watcher
    if (!this._generatedFiles[input]) {
      log$4('Watching ', input);
      watcher.watch(input, {
        change: this._onChange.bind(this),
        unlink: this._onUnlink.bind(this)
      });
    }
    if (!this._actionsByInput[input]) {
      this._actionsByInput[input] = [];
    }
    this._actionsByInput[input].push(action);
    this._actionsByInput[input] = vendor_js.uniq(this._actionsByInput[input]);
  }

  _registerActionOutput (action, output) {
    this._generatedFiles[output] = true;
  }

  _onChange (file) {
    this._invalidate(file);
    const actions = this._actionsByInput[file] || [];
    actions.forEach(function (action) {
      log$4('Rescheduling action', action.id);
      this._schedule(action);
    }.bind(this));
  }

  _onUnlink (file) {
    // TODO: maybe we should unregister the action
    // when the file is deleted
    this._invalidate(file);
  }

  _invalidate (file) {
    log$4('Invalidating', file);
    const actions = this._getAllActions(file);
    actions.forEach(function (action) {
      action.invalidate();
    });
  }

  // find all actions by investigating action input-output dependencies
  _getAllActions (file) {
    let visited = {};
    let queue = [file];
    let actions = [];
    while (queue.length > 0) {
      const next = queue.shift();
      const nextActions = this._actionsByInput[next];
      for (let id in nextActions) {
        if (!nextActions.hasOwnProperty(id)) continue
        if (visited[id]) continue
        const action = nextActions[id];
        visited[id] = true;
        actions.push(action);
        queue = queue.concat(action.outputs);
      }
    }
    return actions
  }

  _startWatching () {
    this.watcher.start();
  }

  _startServing () {
    this._info('Starting server on port %s...', this._serverPort);
    this.server.listen(this._serverPort);
  }

  _schedule (action) {
    const id = action.id;
    const schedule = this._scheduledActions;
    const scheduledIds = this._scheduledActionIds;
    // action is already scheduled
    if (scheduledIds[id]) {
      let idx = -1;
      for (let i = 0; i < schedule.length; i++) {
        if (schedule[i].action.id === id) {
          idx = i;
          break
        }
      }
      if (idx < 0) {
        console.error('Internal Error: tried to reschedule an action with id "%s"', id);
        process.exit(1);
      }
      schedule.splice(idx, 1);
    }

    log$4('Scheduling action: %s', id);
    const entry = { action };
    // return a promise so that you
    // can do things such as b.js().then(...)
    entry.promise = new Promise((resolve, reject) => {
      entry.resolve = resolve;
      entry.reject = reject;
    });
    if (!this._started) {
      schedule.push(entry);
      scheduledIds[id] = true;
    } else {
      process.nextTick(function () {
        schedule.push(entry);
        scheduledIds[id] = true;
        if (!this._running) {
          this._running = true;
          this._next();
        }
      }.bind(this));
    }
    return entry
  }

  _start () {
    log$4('Bundler is starting...');
    this._started = true;
    this._running = true;
    this._step();
  }

  _next () {
    const self = this;
    const entry = this._scheduledActions.shift();
    const action = entry.action;
    const id = action.id;
    const _step = this._step;
    delete this._scheduledActionIds[id];
    try {
      Promise.resolve(action.execute(this))
        .then(_next)
        .catch(_catch);
    } catch (err) {
      _catch(err);
    }

    function _next () {
      const outputs = action.outputs || [];
      outputs.forEach(function (output) {
        const nextActions = self._actionsByInput[output] || [];
        nextActions.forEach(function (action) {
          self._schedule(action);
        });
      });
      entry.resolve();
      // give the system some time if something goes wrong
      // e.g. bundler is in an infinite update look for instance
      setTimeout(_step, 10);
    }

    function _catch (err) {
      log$4('caught exception in action:', err);
      if (err) {
        if (err.stack) {
          console.error(vendor_js.colors.red('\nError during execution of Command '), vendor_js.colors.white(action.name));
          console.error(vendor_js.colors.red(err.toString()));
          console.error('Stacktrace:');
          console.error(err.stack);
        } else {
          console.error(err.toString());
        }
      }
      if (self._firstRun) {
        log$4('... exiting after error on first run');
        process.exit(1);
      } else {
        entry.reject(err);
        _next();
      }
    }
  }

  _step () {
    if (this._scheduledActions.length > 0) {
      this._next();
    } else {
      this._running = false;
      this._firstRun = false;
      this.emit('done');
    }
  }

  _runTasks (names) {
    names.forEach(function (name) {
      this._runTask(name);
    }.bind(this));
  }

  _runTask (name, state) {
    state = state || {};
    if (state[name] === 'done') return
    const task = this._tasks[name];
    if (!task) {
      console.error("Unknown task: '%s'. Exiting", name);
      process.exit();
    }
    state[name] = 'visiting';
    if (task.deps && !this.opts.nodeps) {
      task.deps.forEach(function (dep) {
        if (state[dep] === 'visiting') {
          throw new Error('Cyclic dependency detected in task ' + dep)
        } else if (state[dep] === 'done') ; else {
          this._runTask(dep, state);
        }
      }.bind(this));
    }
    if (task.fn) task.fn();
    state[name] = 'done';
  }

  get _isBundler () { return true }
}

// enabling source maps so that errors in bundler can be traced
require('source-map-support').install();

const log$5 = vendor_js.debug('bundler:main');

const argv = vendor_js.yargs
  // use this to start watcher
  .boolean('h').alias('h', 'help').default('h', false)
  .boolean('w').alias('w', 'watch').default('w', false)
  .boolean('s').alias('s', 'serve').default('s', false)
  .boolean('t').default('t', false)
  .boolean('n').alias('n', 'nodeps').default('n', false)
  .argv;

const showHelp = argv.h;
const showTasks = argv.t;

const bundler = new Bundler({
  watch: argv.watch,
  serve: argv.serve,
  nodeps: argv.nodeps
});
bundler.yargs = vendor_js.yargs;
bundler.argv = argv;
bundler.autorun = true;
const tasks = argv._;

bundler.usage = `
  node make --help
    shows this help

  node make -t
    shows a summary of all available tasks

  node make [-w] [-s] ...<task>
    runs one or multiple tasks and all their dependencies

  Options:
    -w:     watch for file changes
    -s:     start http server
`;
function _start () {
  if (bundler.autorun) {
    if (showHelp) {
      console.info('Usage:');
      console.info(bundler.usage);
      return
    } else if (showTasks) {
      const tasks = bundler._tasks;
      console.info('Available tasks:');
      console.info(Object.keys(tasks).map((name) => {
        const task = tasks[name];
        const frags = ['  - ', vendor_js.colors.green(name)];
        if (task.description) {
          frags.push(vendor_js.colors.white(': '));
          frags.push(task.description);
        }
        return frags.join('')
      }).join('\n'));
      return
    }
    if (tasks.length > 0) {
      log$5('One or more tasks specified. Start running tasks');
      bundler._runTasks(tasks);
    } else if (bundler._tasks['default']) {
      log$5('No tasks specified. Start running default task');
      bundler._runTask('default');
    }
    if (!bundler._hasScheduledActions()) {
      console.error('No action. ');
    }
  }
  bundler._start();
}

process.once('beforeExit', _start);
bundler.once('done', function () {
  log$5('received done...');
  if (showHelp || showTasks || !bundler.autorun) return
  const watch = bundler.opts.watch;
  const serve = bundler.opts.serve;
  const remote = argv.remote;
  if (serve) {
    bundler._startServing();
  }
  if (watch) {
    // start watching if this has been requested (default is false)
    if (!remote) {
      console.info('Watching for changes... Press CTRL-C to exit.');
    }
    bundler._startWatching();
  }
});

if (argv.remote) {
  bundler.once('done', function () {
    process.send('done');
  });
}

module.exports = bundler;
